# 課題レポート4: 力技で解く巡回セールスマン問題

＜目次＞
<ul>
<li><a href="#abst">課題概要</a>
<li><a href="#problem">問題ファイル一式</a>
<li><a href="#definition">巡回経路と巡回経路長の考え方</a>
<li><a href="#goal">課題の達成目標</a>
<li><a href="#detail">課題詳細</a>
<li><a href="#report">取り組み方</a>
<li><a href="#submit">提出方法</a>
</ul>

<hr>

## <a name="abst">課題概要</a>
[巡回セールスマン問題](https://ja.wikipedia.org/wiki/巡回セールスマン問題)とは、与えられた都市集合全てをちょうど一度ずつ巡回し、出発都市に戻る際の総巡回経路長が最小となる経路を求める問題である。今回のレポート課題では「全ての経路長を調べ尽くした上で、最小となる経路を見つける（＝力技）」ようなプログラムを実装すること。なお、経路の組み合わせ数は後述するように「(N-1)!/2」通りとなるため、都市数が多い問題に対しては事実上解けない（調べ尽くすことができない）ことに注意すること。

<hr>

## <a name="problem">問題ファイル一式</a>
- [report4.zip](https://ie.u-ryukyu.ac.jp/~tnal/2016/prog1/report4.zip)　＊準備中
  - 都市配置例
    - input1.txt, input2.txt, input3.txt, input4.txt
      - 1行目に都市数、2行目以降に各都市のx,y座標を記載。（例えば input1.txt の1行目は「8」であり、8都市分のデータが記載されていることを意味している）
    - cities_input1.txt.pdf, cities_input2.txt.pdf, cities_input3.txt.pdf, cities_input4.txt.pdf
  - 都市配置図作成スクリプト
    - plot_map.py
    - 使い方: ``python3 plot_map.py input1.txt``

<hr>

## <a name="definition">巡回経路と巡回経路長の考え方</a>
<img src="https://ie.u-ryukyu.ac.jp/~tnal/2016/prog1/cities_input1.txt.png" width="50%">
- 上図は input1.txt の都市配置図であり、都市0〜都市7までの合計8都市が配置されている。
- セールスマンは都市0から出発し、残りの全て都市を1度ずつ巡回して都市0に戻るものとする。
  - 例えば「都市0->都市1->都市2->...->都市7->都市0」のように巡回する経路を「012345670」と表記するとしよう。この時の経路長は「都市0->都市1」+「都市1->都市2」+...「都市7->都市0」により求められる。
  - 「012345670」は一例であり、可能な巡回経路数は``7!/2``だけ存在する。すなわち都市数をCと置くと``(C-1)!/2``であり、都市数の増加に応じて指数関数的に可能な巡回路数が増えてしまう。
    - 例えば、都市数が3の時の巡回経路は「0120」「0210」の2パターンが考えられるが、この2つの経路は向きを考慮しなければ全く同じ経路であり、事実上1パターンしか存在しないことと同等である。「0120」を後ろから巡回すると「0210」になる。上記式において「/2」で割り算しているのは、向きを無視しているためである。
    - 同様に、都市数が4の時の巡回経路は「01230」「01320」「02130」「02310」「03120」「03210」の6パターンがあるが、向きを無視すると3パターンとなる。
- 実装上は向きを考慮しても無視してもどちらでも構わないが、漏れ無く調べ尽くすようにしよう。
- 可能な巡回路全てを調べつくし、最短となる経路を求めよう。

<hr>

## <a name="goal">課題の達成目標</a>
- 可能な解をすべて列挙し尽くそう。
  - 解の候補が指数関数的に増加する問題の場合、力技で解く（＝可能な解を全て調べ尽くす）のは無駄が多いという意味で本来はよろしくないが、今回はプログラミングの演習として「どうやれば全て調べ尽くすことができるか」を考え、実際に実装してみよう。
- 適度に関数化しよう。
  - 今回の課題は具体的な関数仕様は指定しない。自由に設計・実装してみよう。ただし、1つの関数や、関数になっていないコードが50行を超える場合には分割や関数化すること。
- 自作した関数全てについて、docstringによるドキュメントを書こう。
- 自作した関数全てについて、doctestによるユニットテストを書こう。

<hr>

## <a name="detail">課題詳細</a>
1. ダウンロードしたファイルには4種類の都市配置（input1.txt〜input4.txt）を用意している。これら4つの都市配置における最短路となる巡回経路を、力技で求めるプログラムを作ろう。レポートにはinput1.txt〜input4.txtの4つを各々処理した結果を掲載すること。
  - 実行イメージ（下記）
    - 厳密に下記通りの出力とならなくても良いが、<font color="red">(1)読み込んだ都市数、(2)都市座標、(3)求めた最短経路長、(4)その際の巡回経路の4点を実行結果に含めること</font>。
    - 下記実行イメージではコマンドラインにて入力ファイルを指定している。この方法は
2. 4種類の都市配置（input1.txt〜input4.txt）各々について、最単純回路を求めるのに要した時間を計測し、「横軸に都市数、縦軸に実行時間(sec)」を取る折れ線グラフを作図せよ。
  - 時間計測には「time」コマンドを用いよ。例えばターミナル上で「time python3 tsp.py input4.txt」として実行すると、処理に要した時間を計測し、出力してくれる。
    - 時間計測結果の読み方
      - 「python3 tsp.py input4.txt  23.31s user 0.49s system 98% cpu 24.186 total」のように出力された場合、トータルで24.186秒かかったことを意味する。
  - グラフ作成には gnuplot か matplotlib を用いること。
    - gnuplotはソフトウェア演習で演習済み。
    - matplotlibは、ダウンロードファイルに含まれる「plot_map.py」を参考にするか、別途授業で補足する予定です。
3. 4種類の都市配置（input1.txt〜input4.txt）を解くのに要した時間から、都市数が12〜20まで増えたと仮定した時の実行時間を見積もれ。また、見積もった時間から、力技で解くアプローチの特徴について考察せよ。
  - 「見積もった時間」は目安で構わないが、理由をつけて見積もること。（どれだけ精度高い見積もりを出したとしても、理由が記述されていなければ減点です）
  - ヒント
    - input1.txtは8都市、input2.txtは9都市、input3.txtは10都市、input4.txtは11都市の例である。全巡回経路数は都市数により定まり、都市が増えると指数関数的に経路数が増える。経路数の増え具合と、実行時間との増え具合の間に何かしらパターンを見出すことはできないだろうか？
4. 自作した関数全てについて docstringによるドキュメント、doctestによるテストを記述せよ。
  - docstringドキュメントには、(a)関数1行概要、(b)引数説明、(c)戻り値説明、の3点を記述すること。
    - 参考: [レポート2](https://ie.u-ryukyu.ac.jp/~tnal/2016/prog1/report2.html)
5. <font color="red">(オマケ)</font> 余裕がある人向けの発展課題です。前述1では力技で最短経路を見つけたが、「最適でなくても構わないので、ある程度良質の解を、低コストで見つける（なるべく無駄を省く）」ことはできないだろうか？　自由に考え、提案し、実装してみよ。
  - レポートには締め切りがある。締め切りに間に合わない場合には、取り組み具合が分かるように報告すること。（報告がない場合には加点しようがないです）

```
# 実行イメージ
% python3 report4.py input1.txt
city_num = 8
[[20, 20], [120, 20], [220, 20], [70, 120], [170, 120], [270, 120], [20, 220], [120, 220]]
min_dist = 847.2135954999579, min_path = [0, 3, 6, 7, 4, 5, 2, 1, 0]

% python3 report4.py input2.txt
city_num = 9
[[20, 20], [120, 20], [220, 20], [70, 120], [170, 120], [270, 120], [20, 220], [120, 220], [220, 220]]
min_dist = 947.2135954999579, min_path = [0, 3, 6, 7, 8, 4, 5, 2, 1, 0]

% python3 tsp.py input3.txt
city_num = 10
[[20, 20], [120, 20], [220, 20], [320, 20], [70, 120], [170, 120], [270, 120], [20, 220], [120, 220], [220, 220]]
362880
min_dist = 1047.213595499958, min_path = [0, 1, 2, 3, 6, 5, 9, 8, 7, 4, 0]

% python3 tsp.py input4.txt
city_num = 11
[[20, 20], [120, 20], [220, 20], [320, 20], [70, 120], [170, 120], [270, 120], [370, 120], [20, 220], [120, 220], [220, 220]]
3628800
min_dist = 1147.2135954999578, min_path = [0, 1, 2, 3, 7, 6, 5, 10, 9, 8, 4, 0]
```

<hr>

## <a name="report">取り組み方</a>
- ペアや友人らと話し合って取り組んで構わないが、コード解説を加えるなど「自分自身の報告書」となるように取り組むこと。試して分かったこと、自身で解決できなかった部分等についてどう取り組んだか、といった過程がわかるように示すこと。（考えを図表や文章を駆使して表現して報告する練習です）
- レポート作成は好きなツール（ソフトウェア）を使って構わない。ただし下記を含めること。
  - タイトル
    - 今回は「**プログラミング1、レポート課題4: 「力技で解く巡回セールスマン問題」**」。
  - 提出日: yyyy-mm-dd
  - 報告者: 学籍番号、氏名
    - 複数人で相談しながらやった場合、相談者らを「**協力者: 学籍番号、氏名**」として示そう。
  - 課題説明
  - 結果と考察
    - **課題への取り組みを通し、課題の意義、課題から分かったこと、今後の展望などを述べる。失敗やつまづきがあれば、それらについての失敗分析を含めること。**
    - 参考リンク: [実験レポートの書き方](http://www.report.gusoku.net/jikken/jikkenreport.html)
  - その他
    - 通常は感想等をレポートには含めませんが、練習なので課題に取り組みながら何か感じたこと、悩んでいること等、書きたいことがあれば自由に書いてください。（なければ省略OK）

<hr>

## <a name="submit">提出方法</a>
- 提出物は「レポート」、「作成したスクリプトファイル」の2点である。
- レポートは電子ファイルで提出するものとする。
- 提出先:
  - 「<a href="https://drive.google.com/a/ie.u-ryukyu.ac.jp/folderview?id=0B8oAeomiuJo-RkcyVlNmLWdCcFE&usp=sharing">Google ドキュメント</a>」のreport4。
- 締切: 8/4(木)。
